/*
 * @(#)Process.cpp
 *
 * This file is part of webCDwriter - Network CD/DVD Writing.
 *
 * Copyright (C) 1999-2008 Jrg P. M. Haeger
 *
 * webCDwriter is free software. See CDWserver.cpp for details.
 *
 * Jrg Haeger, 02.11.1999
 */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>
#include <string.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>
#include "Config.h"
#include "Support.h"
#include "LineReader.h"
#include "Log.h"
#include "Process.h"

Process::Process(const char *fileName) {
	init(fileName);
}

Process::~Process() {
	if (pid > 0 && result < 0) {
		kill(pid, SIGKILL);
		getResult();
	}
	log.debug("Process", "~Process() %s %d result = %d",
		args[0], pid, result);

	if (protoStream != NULL && protoStream != stdout)
		fclose(protoStream);
	if (inErrorStream != NULL)
		delete inErrorStream;
	for (int i = 0; i < envsNum; i++)
		delete[] envs[i];
	for (int i = 0; i < argsNum; i++)
		delete[] args[i];
	closePipes();
}

void Process::addArg(const char *format, ...) {
	if (argsNum == sizeof args - 1)
		throw new Exception("Process::addArg: to many args");

	char str[1024];
	va_list ap;
	va_start(ap, format);
	vsnprintf(str, sizeof str, format, ap);
	va_end(ap);

	args[argsNum++] = strDup(str);
	args[argsNum] = NULL;
}

void Process::addArg(String &str) {
	try {
		addArg("%s", str.getBytes());
	}
	catch (Exception *e) {
		String::free(&str);
		throw e;
	}
	String::free(&str);
}

void Process::addArgs(const char *args) {
	char str[strlen(args) + 1];
	for (int i = 0; ; i++, args++) {
		str[i] = *args;
		if (str[i] == 0 || str[i] == ' ' || str[i] == '\t') {
			str[i] = 0;
			if (i > 0)
				addArg(str);
			i = -1;
		}
		if (*args == 0)
			break;
	}
}

void Process::addEnv(const char *format, ...) {
	if (envsNum == sizeof envs - 1)
		throw new Exception("Process::addEnv: to many envs");

	char str[1024];
	va_list ap;
	va_start(ap, format);
	vsnprintf(str, sizeof str, format, ap);
	va_end(ap);

	envs[envsNum++] = strDup(str);
	envs[envsNum] = NULL;
}

void Process::closePipes() {
	if (toChildStream >= 0) {
		close(toChildStream);
		toChildStream = -1;
	}

	if (fromChildStream >= 0) {
		close(fromChildStream);
		fromChildStream = -1;
	}

	if (errorFromChildStream >= 0) {
		close(errorFromChildStream);
		errorFromChildStream = -1;
	}
}

void Process::closePipeToChild() {
	if (toChildStream >= 0) {
		close(toChildStream);
		toChildStream = -1;
	}
}

int Process::getErrorStream() {
	return errorFromChildStream;
}

int Process::getInputStream() {
	return fromChildStream;
}

int Process::getOutputStream() {
	return toChildStream;
}

int Process::getResult() {
	log.debug("Process", "getResult() %d %d", pid, result);
	if (pid <= 0 || result >= 0)
		return result;

	if (inErrorStream != NULL)
		delete inErrorStream;
	inErrorStream = NULL;

	if (inStream != NULL)
		delete inStream;
	inStream = NULL;

	int res, status;
//	TEMP_FAILURE_RETRY(waitpid(pid, &status, 0));
	while ((res = waitpid(pid, &status, 0)) == -1 && errno == EINTR);

	if (res == -1)
		result = 256;
	else
	if (WIFEXITED(status))
		result = WEXITSTATUS(status);
	else
	if (WIFSIGNALED(status))
		result = 1000 + WTERMSIG(status);
	else
		result = 257;
	return result;
}

void Process::init(const char *fileName) {
	argsNum = envsNum = 0;
	args[argsNum++] = strDup(fileName);
	args[argsNum] = NULL;
	envs[envsNum] = NULL;
	pid = -1;
	result = -1;
	toChildStream = -1;
	fromChildStream = -1;
	errorFromChildStream = -1;
	fromParentStream = -1;
	inErrorStream = NULL;
	inStream = NULL;
	protoStream = NULL;
}

int Process::isLineAvailable() {
	return (inErrorStream != NULL && inErrorStream->available())
		|| (inStream != NULL && inStream->available());
}

int Process::read(char *buf, int count) {
	fcntl(fromChildStream, F_SETFL, O_NONBLOCK);
	return ::read(fromChildStream, buf, count);
}

/**
 * A log filter for readcd.
 */
inline bool filter(const char *line) {
	int addr;
	return sscanf(line, "addr: %d", &addr) == 1 && addr % 5120 != 0;
}

const char *Process::readLine(int timeout) {
	const char *lineStr = NULL;
	if (inErrorStream == NULL)
		;
	else
	if (inStream == NULL || inErrorStream->available())
		lineStr = inErrorStream->next(timeout);
	else
	if (inStream->available())
		lineStr = inStream->next(timeout);
	else {
		fd_set fdSet;
		struct timeval tv;
		if (timeout > 0) {
			FD_ZERO(&fdSet);
			FD_SET(fromChildStream, &fdSet);
			FD_SET(errorFromChildStream, &fdSet);
			tv.tv_sec = timeout;
			tv.tv_usec = 0;
			if (::select(FD_SETSIZE, &fdSet, NULL, NULL, &tv) == 0)
				throw new Exception("timeout");
		}
		else {
			while (true) {
				FD_ZERO(&fdSet);
				FD_SET(fromChildStream, &fdSet);
				FD_SET(errorFromChildStream, &fdSet);
				tv.tv_sec = 1;
				tv.tv_usec = 0;
				if (::select(FD_SETSIZE, &fdSet, NULL, NULL, &tv) > 0)
					break;
			}
		}
		if (FD_ISSET(errorFromChildStream, &fdSet))
			lineStr = inErrorStream->next(timeout);
		if (lineStr == NULL)
			lineStr = inStream->next(timeout);
	}
	if (lineStr != NULL && !filter(lineStr))
		if (protoStream != NULL)
			if (protoStream != stdout) {
				fprintf(protoStream, "%s\n", lineStr);
				fflush(protoStream);
			}
			else
			if (!filter(lineStr))
				log.put(5, S.e + "S"
					+ (pid < 10000? "0": "")
					+ (pid < 1000? "0": "")
					+ (pid < 100? "0": "")
					+ (pid < 10? "0": "")
					+ pid + " " + lineStr);
	return lineStr;
}

int Process::run() {
	start();
	while (1) {
		const char *lineStr = readLine();
		if (lineStr == NULL)
			break;
	}
	return getResult();
}

void Process::select() {
	fd_set fdSet;
	FD_ZERO(&fdSet);
	if (fromChildStream >= 0)
		FD_SET(fromChildStream, &fdSet);
	if (errorFromChildStream >= 0)
		FD_SET(errorFromChildStream, &fdSet);

	fd_set fdWriteSet;
	FD_ZERO(&fdWriteSet);
	if (toChildStream >= 0)
		FD_SET(toChildStream, &fdWriteSet);

	::select(FD_SETSIZE, &fdSet, &fdWriteSet, NULL, NULL);
}

void Process::sendTerminateSignal() {
	if (pid > 0 && result < 0) {
		log.put(5, S.e + "kill " + pid);
		int res = kill(pid, SIGTERM);
		log.put(5, S.e + "kill " + pid + " = " + res);
		if (res != 0) {
			log.put(5, S.e + "kill " + errno + " " + strerror(errno));
			String name = "/tmp/CDWcancel-";
			name = name + pid;
			FILE *out = fopen(name.getBytes(), "w");
			if (out != NULL)
				fclose(out);
		}
	}
}

void Process::setFromParentStream(int fd) {
	fromParentStream = fd;
}

void Process::setProtoFile(const char *fileName) {
	if (fileName != NULL && config.getLogLevel() > 9) {
		makeDirs(fileName);
		protoStream = fopen(fileName, "w");
		if (protoStream == NULL)
			protoStream = stdout;
	}
	else
		protoStream = stdout;
}

void Process::show() {
	log.put(4, S.e + "Process: \"" + args[0] + "\"");
	for (int i = 1; i < argsNum; i++)
		log.put(4, S.e + "      " + (i < 10? " ": "") + i
			+ ": \"" + args[i] + "\"");
}

/**
 *@see glibc-2.1.1/sysdeps/posix/pipestream.c
 */
void Process::start(const char *workDirName, int splitOut) {
	int outPipe[2], inPipe[2], errPipe[2];
	if (fromParentStream < 0) {
		if (pipe(outPipe) != 0)
			throw new Exception("outPipe failed");
	}
	else {
		outPipe[0] = fromParentStream;
		outPipe[1] = -1;
	}
	if (pipe(inPipe) != 0)
		throw new Exception("inPipe failed");
	if (pipe(errPipe) != 0)
		throw new Exception("errPipe failed");

	int errorToParentStream = errPipe[1];
	errorFromChildStream = errPipe[0];
	int toParentStream = inPipe[1];
	fromChildStream = inPipe[0];
	toChildStream = outPipe[1];
	fromParentStream = outPipe[0];

	bool debugProcessStart
		= config.getDebug().indexOf("Process::start") >= 0;

	char logDir[1024];
	strncpy(logDir, config.getLogDir(), sizeof logDir - 1);
	logDir[sizeof logDir - 1] = 0;

	pid = fork();
	if (pid == (pid_t)-1)
		throw new Exception("fork failed");

	if (pid == (pid_t)0) {
		// We are the child side.
		FILE *logFile = NULL;
		if (debugProcessStart) {
			char logFileName[1024];
			sprintf(logFileName, "%s/childs/%d",
				logDir, getpid());
			makeDirs(logFileName);
			logFile = fopen(logFileName, "w");
			if (logFile != NULL) {
				fprintf(logFile, "1\n");
				fflush(logFile);
			}
		}

		dup2(fromParentStream, STDIN_FILENO);
		dup2(toParentStream, STDOUT_FILENO);
		dup2(errorToParentStream, STDERR_FILENO);

		if (logFile != NULL) {
			fprintf(logFile, "2\n");
			fflush(logFile);
		}

		closePipes();
		close(fromParentStream);
		close(toParentStream);
		close(errorToParentStream);

		if (logFile != NULL) {
			fprintf(logFile, "3\n");
			fflush(logFile);
		}

		if (workDirName != NULL)
			if (chdir(workDirName) != 0)
				exit(129);

		if (logFile != NULL) {
			fprintf(logFile, "4\n");
			fflush(logFile);
		}

		if (envsNum == 0)
			execv(args[0], args);
		else
			execve(args[0], args, envs);

		if (logFile != NULL) {
			fprintf(logFile, "5\n");
			fclose(logFile);
		}

		exit(128);
	}

	log.debug("Process", "start() %s %d", args[0], pid);
	close(fromParentStream);
	fromParentStream = -1;
	close(toParentStream);
	toParentStream = -1;
	close(errorToParentStream);
	errorToParentStream = -1;

	if (splitOut) {
		inErrorStream = new LineReader(errorFromChildStream);
		inStream = NULL;
//		errorFromChildStream = -1;
	}
	else {
		inErrorStream = new LineReader(errorFromChildStream);
		inStream = new LineReader(fromChildStream);
//		fromChildStream = -1;
//		errorFromChildStream = -1;
	}
	if (inErrorStream == NULL)
		throw new Exception("open inErrorStream failed");
}

void Process::terminate() {
	if (pid > 0 && result < 0) {
		kill(pid, SIGTERM);
		getResult();
	}
}

const char *Process::toString(char *str) {
	sprintf(str, "Process: <%s", args[0]);
	for (int i = 1; i < argsNum; i++) {
		strcat(str, " ");
		strcat(str, args[i]);
	}
	strcat(str, ">");
	return str;
}

int Process::write(const char *buf, int count) {
	fcntl(toChildStream, F_SETFL, O_NONBLOCK);
	return ::write(toChildStream, buf, count);
}
